/*----------------------------------------------------------------------
  File    : numsort.c
  Contents: several algorithms for sorting number arrays
  Author  : Christian Borgelt
  History : 1998.01.21 file created
            1998.02.28 quicksort and combsort added
            1998.03.03 shellsort, heapsort, and mergesort added
            1998.04.02 binsort added
            1998.04.03 some comments added
            1998.05.22 more comments added
            1998.07.15 minor improvements
            1998.09.16 determination of median improved
            2013.03.06 smoothsort added
            2013.09.30 bug in function leosift() fixed
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <float.h>
#include <math.h>
#include <time.h>

#ifndef CHECK
#define CHECK   0               /* whether to check the sort result */
#endif

#ifndef SIZE_FMT
#  ifdef _MSC_VER
#  define SIZE_FMT  "Iu"        /* printf format code for size_t */
#  else
#  define SIZE_FMT  "zu"        /* printf format code for size_t */
#  endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE   65536         /* block size for number array */
#define TH_INSERT    16         /* threshold for insertion (combsort) */

/* --- error codes --- */
#define OK            0         /* no error */
#define E_NONE        0         /* no error */
#define E_NOMEM     (-1)        /* not enough memory */
#define E_FOPEN     (-2)        /* file open failed */
#define E_FREAD     (-3)        /* file read failed */
#define E_FWRITE    (-4)        /* file write failed */
#define E_METHOD    (-5)        /* illegal sort method */
#define E_UNKNOWN   (-6)        /* unknown error */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error\n",
  /* E_NOMEM    -1 */  "not enough memory\n",
  /* E_FOPEN    -2 */  "cannot open file `%s'\n",
  /* E_FREAD    -3 */  "read error on file `%s'\n",
  /* E_FWRITE   -4 */  "write error on file `%s'\n",
  /* E_METHOD   -5 */  "illegal sort method %i\n",
  /* E_UNKNOWN  -6 */  "unknown error\n"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static char *prgname;           /* program name for error messages */

/*----------------------------------------------------------------------
  Auxiliary Function
----------------------------------------------------------------------*/

static int fltcmp (const void *p1, const void *p2)
{                               /* --- compare two float numbers */
  if (*(const float*)p1 > *(const float*)p2) return  1;
  if (*(const float*)p1 < *(const float*)p2) return -1;
  return 0;                     /* return sign of *p1 -*p2 */
}  /* fltcmp() */

/*----------------------------------------------------------------------
The above function is used in the call to the standard library function
`qsort', which needs a comparison function to sort an array of arbitrary
objects. The comparison function determines the order of the objects in
the sorted array. With the above function the numbers are sorted in
ascending order. If the 1 and -1 in the two `return' statements are
exchanged, the array is sorted in descending order.
----------------------------------------------------------------------*/
/*----------------------------------------------------------------------
  Sort Functions
----------------------------------------------------------------------*/

void bubblesort (float a[], size_t n)
{                               /* --- bubblesort */
  size_t i;                     /* loop variable, index */
  int    exg;                   /* whether numbers were exchanged */
  float  t;                     /* exchange buffer */

  do {                          /* sort loop */
    exg = 0;                    /* clear exchanged flag */
    n--;                        /* get index of last element */
    for (i = 0; ++i <= n; ) {   /* traverse the number array */
      if (a[i-1] > a[i]) {      /* if numbers are in wrong order */
        t = a[i-1]; a[i-1] = a[i]; a[i] = t; exg = 1; }
    }                           /* exchange numbers and set flag */
  } while (exg);                /* while numbers were exchanged */
}  /* bubblesort() */

/*----------------------------------------------------------------------
Bubblesort is one of the simplest sorting algorithms. It repeatedly
traverses the data array and looks for pairs of objects (here: numbers)
that are in wrong order. If such a pair is found, the objects (numbers)
are exchanged. The algorithm stops, if all objects are in correct order,
i.e. if no pairs had to be exchanged during a traversal of the array.
  The name `bubblesort' stems from the fact that with this method the
large objects (numbers) ascend like bubbles in the data array. For
example, consider the array [5 3 1 4 2]. The first traversal of this
array lets the 5 ascend in the array:

  5 3 1 4 2        initial situation
  3 5 1 4 2        5 and 3 exchanged
  3 1 5 4 2        5 and 1 exchanged
  3 1 4 5 2        5 and 4 exchanged
  3 1 4 2 5        5 and 2 exchanged

Note that roughly the same would have happened, if the 5 had not been
the first element of the array, but, say, the third. In this case the
5 would have been ``picked up'' when the algorithm had reached the
third element and then the 5 would have ascended in the array in the
same manner as above. In any case, after one traversal of the array
the largest number is stored in the last array element. Hence, in
subsequent traversals, the array need not be traversed up to its
upper bound, since we already know that the largest number is last.
This explains the `n--' in the implementation above, which reduces the
part of the array traversed.
  In order to determine when the algorithm has completed its task, the
above function uses a flag `exg', which indicates whether two objects
(numbers) where exchanged during the last traversal of the array. It
is set to false before traversing the array and to true, if two objects
are exchanged. Thus, if `exg' is false when execution reaches the last
line of the above function (the `while' part of the do-while-statement),
we know that no pair of objects has been exchanged and hence the array
must be sorted. This leads to a favorable property of this algorithm:
if the array is already sorted, it stops after one traversal.

Possible improvements:
Since this algorithm lets the greatest object ``picked up'' ascend
as far as necessary in one go, we know that the array must already
be sorted beyond the highest position an exchange was made. Were it
otherwise, a number had ascended beyond that position. Hence we can
sometimes restrict the part of the array to be traversed more severely
than by just reducing the upper limit by one in each step. In addition,
we can traverse the array alternatingly upwards and downwards. During
an upward traversal the greatest object picked up ascends as far as
necessary, during a downward traversal the least object picked up
descends as far as necessary. By the same argument as above we can then
raise the lower limit of the part of the array to be traversed to the
lowest position where an exchange was made.

Exercises:
1. What is the time complexity of the bubblesort algorithm?
2. Implement the two improvements mentioned above!
3. Does the time complexity change due to the improvements?
----------------------------------------------------------------------*/

void selsort (float a[], size_t n)
{                               /* --- selection sort */
  size_t i, k;                  /* loop variables and indices */
  float  t;                     /* exchange buffer */

  while (--n > 0) {             /* do n-1 selections */
    for (i = k = n; i > 0; )    /* traverse the unsorted numbers and */
      if (a[--i] > a[k]) k = i; /* find index of the greatest number */
    t = a[n]; a[n] = a[k]; a[k] = t;
  }                             /* exchange the greatest number */
}  /* selsort() */              /* and the current last number */

/*----------------------------------------------------------------------
Selection sort is equally simple as bubblesort. The idea underlying
this algorithm is to traverse the array and to determine the greatest
object (here: number) in it. This object (number) is swapped to the last
element of the array (i.e. it is exchanged with the object currently
occupying that position). The selection process is then repeated with
an array one element smaller (i.e. excluding the last element). If the
remaining array contains only one object (number), the whole array is
sorted.
  As an example, consider the array [5 3 1 4 2] and trace how it is
sorted. First the 5 is selected and swapped to the last place, then
the 4 is selected and swapped to the last place but one, and so on.
The part of the array from which the largest element is selected is
separated from the rest by a `|'.

step   state of the array    greatest element
1)     5 3 1 4 2|               5
2)     2 3 1 4|5                4
3)     2 3 1|4 5                3
4)     2 1|3 4 5                2
5)     1|2 3 4 5

Exercise:
What is the time complexity of the selection sort algorithm?
----------------------------------------------------------------------*/

void inssort (float a[], size_t n)
{                               /* --- insertion sort */
  size_t i, k;                  /* loop variables, indices */
  float  t;                     /* exchange buffer */

  for (i = 0; ++i < n; ) {      /* traverse numbers to insert */
    t = a[i];                   /* note the number to insert */
    for (k = i; a[k-1] > t; ) { /* traverse the numbers to the left */
      a[k] = a[k-1];            /* shift right all numbers that */
      if (--k <= 0) break;      /* that are greater than */
    }                           /* the number to insert */
    a[k] = t;                   /* store the number to insert */
  }                             /* in the place thus found */
}  /* inssort() */

/*----------------------------------------------------------------------
Insertion sort is also a very simple algorithm, just like bubblesort and
selection sort. The idea underlying it is the following: The array to
be sorted is divided into two sections, one already sorted and one
holding the numbers to be inserted into the sorted section. In each step
one number is removed from the unsorted section, its place within the
sorted section is determined, and then it is stored at that position.
The algorithm stops if the section containing the numbers to insert
becomes empty.
  Consider the array [5 3 1 4 2] and trace how it is sorted. We mark
the dividing line between the sorted section (left part of the array)
and the one holding the numbers to insert (right part of the array)
with a `|'. The number to insert is temporarily removed from the array
and thus one element gets empty. This element is marked by a `*'. In the
beginning the part of the array that consists only of the first element
is obviously sorted, hence we start by inserting the second element.
The indentation used in the trace below represents whether we are in
the inner or the outer loop of the above function.

step   current state    element      step   current state    element
       of the array     to insert           of the array     to insert
1)     5|3 1 4 2          3          3a)       1 3 5 *|2       4
1a)       5 *|1 4 2       3          3b)       1 3 * 5|2       4
1b)       * 5|1 4 2       3          4)     1 3 4 5|2          2
2)     3 5|1 4 2          1          4a)       1 3 4 5 *|      2
2a)       3 5 *|4 2       1          4b)       1 3 4 * 5|      2
2b)       3 * 5|4 2       1          4c)       1 3 * 4 5|      2
2c)       * 3 5|4 2       1          4d)       1 * 3 4 5|      2
3)     1 3 5|4 2          4          5)     1 2 3 4 5|

Possible improvement:
The place to insert the current object (number) at is determined by a
linear search through the objects (numbers) to the left of it. But since
we know that the objects (numbers) to the left of the current object
(number) are sorted, we can do better: we can use a binary search.
Although we still have to shift all numbers to the right of the position
at which to insert the number, this reduces the number of comparisons.

Exercises:
1. What is the time complexity of the selection sort algorithm?
2. Implement the improvement mentioned above!
3. Does the time complexity change due to the improvement?
----------------------------------------------------------------------*/

void shellsort (float a[], size_t n)
{                               /* --- Shell sort */
  size_t i, k;                  /* loop variables, indices */
  size_t h;                     /* increment for index */
  float  t;                     /* exchange buffer */

  for (h = n; h > 1; ) {        /* traverse increments */
    h = (h < 5) ? 1 : ((h *5 -1) /11);
    for (i = h-1; ++i < n; ) {  /* insertion sort with increment h */
      t = a[i];                 /* note the element to insert */
      for (k = i; a[k-h] > t; ){/* traverse numbers to the left */
        a[k] = a[k-h];          /* shift right all numbers that */
        if ((k -= h) < h) break;/* are greater than the number */
      }                         /* to insert and go to the left */
      a[k] = t;                 /* store the number to insert */
    }                           /* in the place thus found */
  }
}  /* shellsort() */

/*----------------------------------------------------------------------
Reference: D.L. Shell
           A High-speed Sorting Procedure
           Communications of the ACM 2(7), 30-32, 1959

Shellsort is basically an insertion sort on subsequences with decreasing
increment. That is, in the first go (with increment i1) the subsequences
1, i1+1, 2*i1+1, etc., 2, i1+2, 2*i1+2, etc. are sorted with a standard
insertion sort. In the next go a new increment i2, smaller than the
first, is employed and the subsequences 1, i2+1, 2*i2+1, etc., 2, i2+2,
2*i2+2, etc. are sorted with standard insertion sort, and so forth.
  The advantage of this method is that numbers that are far from their
final position in the sorted array can ``travel faster'' to this
position, since not all numbers lying in between their current and
their final position have to be moved. Consider, for example, the
following array and let's start with i1 = 4:
  3 6 5 4 1 2
If this array were sorted with standard insertion sort, to move the
numbers 1 and 2 to their correct position would take eight shifts of
numbers. (In order to move the 1 to the front, the 3, 4, 5, and 6 have
to be shifted right one position. The same holds for the 2.) With the
shellsort algorithm it takes only two exchanges.
  The best sequence of increments is not easy to determine. D.E. Knuth
recommends as increments 1, 4, 13, 40, etc. (in reverse order, of
course), i.e. n_{i+1} = 3 *n_i +1. This can be implemented by writing

  for (h = 1; h < n/9; h = 3*h +1);
  for ( ; h > 0; h /= 3) {

(the first loop determines the initial increment) instead of

  for (h = n; h > 1; ) {
    h = (h < 5) ? 1 : ((h *5 -1) /11);

but empirical tests show that it is slower than the version used above.
  We do not ask you to determine the time complexity of this algorithm,
since this is a pretty difficult task that involves a lot of complicated
mathematics. For the sequence of increments recommended by D.E. Knuth
it can be shown that the average time complexity is roughly O(n^1.2).
----------------------------------------------------------------------*/

void mergesort (float a[], size_t n)
{                               /* --- mergesort */
  float  *buf, *b;              /* work buffer */
  float  *t;                    /* to exchange the arrays */
  size_t left;                  /* number of elements left */
  size_t len;                   /* length of sequences */
  size_t cnt1, cnt2;            /* number of source elements */
  size_t s1,   s2;              /* indices for sources */
  size_t d;                     /* index for destination */

  if (n <= 1) return;           /* do not sort less than two elements */
  buf = b = (float*)malloc(n *sizeof(float));
  if (!buf) return;             /* allocate a work buffer */
  len = 1;                      /* start with isolated elements */
  do {                          /* sort loop */
    left = n;                   /* all elements have to be merged */
    s1 = d = 0; s2 = len;       /* get source and destination indices */
    do {                        /* merge loop */
      left -= cnt1 = (left >= len) ? len : left;  /* get numbers */
      left -= cnt2 = (left >= len) ? len : left;  /* of elements */
      while ((cnt1 > 0)         /* while both source sequences */
      &&     (cnt2 > 0)) {      /* are not empty */
        if (a[s1] < a[s2]) {    /* if the next element of the */
          b[d++] = a[s1++];     /* first sequence is smaller, */
          cnt1--; }             /* copy it to the destination */
        else {                  /* if the next element of the */
          b[d++] = a[s2++];     /* second sequence is smaller, */
          cnt2--;               /* copy it to the destination */
        }                       /* (in doing so increment indices */
      }                         /* and decrement element counter) */
      while (cnt1-- > 0)        /* copy remaining elements */
        b[d++] = a[s1++];       /* of first source sequence */
      while (cnt2-- > 0)        /* copy remaining elements */
        b[d++] = a[s2++];       /* of second source sequence */
      s1 += len; s2 += len;     /* get next start indices */
    } while (left > 0);         /* while there are elements left */
    t = a; a = b; b = t;        /* exchange source and destination */
    len <<= 1;                  /* and double sequence length */
  } while (len < n);            /* while the array is not sorted */
  if (a == buf)                 /* if the result is in the buffer, */
    for ( ; n > 0; n--) b[n] = a[n];   /* copy it back to argument */
  free(buf);                    /* delete the work buffer */
}  /* mergesort() */

/*----------------------------------------------------------------------
This is an iterative version of the mergesort algorithm. A recursive
version has been discussed in the lecture. The idea of the mergesort
algorithm is to sort a given sequence by merging sorted subsequences of
increasing length. In the beginning the length of each subsequence is 1,
since the sequence is unsorted. These subsequences of length 1 are
merged into subsequences of length 2, these in turn into subsequences
of length 4 and so on, until the complete array is sorted. For example
(The character '|' separates subsequences):

  5|3|1|4|2
  3 5|1 4|2
  1 3 4 5|2
  1 2 3 4 5

To illustrate the merging of two sequences, consider the situation
with subsequences of length two. To merge the sequences 3 5 and 1 4,
the first elements of these two input sequences are compared and the
lesser element is moved to the output sequence. This is repeated until
one of the input sequences becomes empty. Then the elements of the
remaining (non-empty) sequence are appended to the output list:

  output     input 1   input 2    elements compared   lesser element
             3 5       1 4        3 and 1             1
  1          3 5       4          3 and 4             3
  1 3        5         4          5 and 4             4
  1 3 4      5                    ---                 -
  1 3 4 5

A serious drawback of the mergesort algorithm, when applied to arrays,
is that a buffer as large as the array to sort is neccessary. All other
algorithms do without such additional storage. But this drawback only
applies for arrays. If mergesort is used to sort lists, it does not
need any additional storage (since the merging is done by changing the
list pointers) and thus is a very efficient algorithm.

Exercises:
1. What is the time complexity of the mergesort algorithm?
2. Implement a recursive version of the mergesort algorithm!
----------------------------------------------------------------------*/

static void sift (float a[], size_t lft, size_t rgt)
{                               /* --- let element sift down in heap */
  size_t i;                     /* index of first successor in heap */
  float  t;                     /* buffer for element */

  t = a[lft];                   /* note sift element */
  i = lft +lft +1;              /* compute index of first successor */
  do {                          /* sift loop */
    if ((i    < rgt)            /* if second successor exists */
    &&  (a[i] < a[i+1]))        /* and it is greater than the first, */
      i++;                      /* go to the second successor */
    if (t >= a[i])              /* if the successor is greater */
      break;                    /* than the sift element, */
    a[lft] = a[i];              /* let the successor ascend in heap */
    lft = i;                    /* note index of current element */
    i  += i +1;                 /* compute index of first successor */
  } while (i <= rgt);           /* while still within heap */
  a[lft] = t;                   /* store the sift element */
}  /* sift() */

/*----------------------------------------------------------------------
The above function lets the element at index `lft' sift down in the
array `a' to (re)establish heap structure in the section from index
`lft' to index `rgt'. For a more detailed description see below.
----------------------------------------------------------------------*/

void heapsort (float a[], size_t n)
{                               /* --- heapsort */
  size_t lft, rgt;              /* boundaries of heap section */
  float  t;                     /* exchange buffer */

  if (n <= 1) return;           /* do not sort less than two elements */
  lft = n >> 1;                 /* at start, only the second half */
  rgt = n -1;                   /* of the array has heap structure */
  while (lft > 0)               /* while the heap is not complete, */
    sift(a, --lft, rgt);        /* extend it by one element */
  while (1) {                   /* heap reduction loop */
    t = a[0]; a[0] = a[rgt];    /* swap the greatest element */
    a[rgt] = t;                 /* to the end of the array */
    if (--rgt <= 0) break;      /* if the heap is empty, abort */
    sift(a, 0, rgt);            /* let the element that has been */
  }                             /* swapped to front sift down */
}  /* heapsort() */

/*----------------------------------------------------------------------
An array with n elements is said to have heap structure, if for all i,
0 <= i < n/2-1, it is a[i] >= a[2*i+1] and a[i] >= a[2*i+2], and, if n
is even, then also a[n/2-1] >= a[n-1]. In this case, the array can be
seen as representing a binary tree, in which a[2*i+1] is the first (or
left) successor of a[i] and a[2*i+2] is the second (or right) successor
of a[i]. The heap condition then says that the object in a node has to
be greater than the objects in both of its descendants. See the
diagrams below, in which the nodes are labeled with the indices of the
corresponding array elements (left tree) and, for the number array
[7 6 4 1 5 2 3], with the numbers in these array elements (right tree).

indices:        0             numbers:          7
              /   \                           /   \
            1       2                       6       4
           / \     / \                     / \     / \
          3   4   5   6                   1   5   2   3

In this tree a node is always greater than its two descendants and
thus (since all descendants are in turn greater than their descendants)
greater than all nodes in the subtree of which it is the root. Hence,
if an array has heap structure, it is very simple to determine its
largest element: it must be the first element of the array.
  Remember that the selection sort algorithm worked by determining the
largest object (number) in the unsorted part of the array and swapping
this object (number) to the end of the array. The heap sort algorithm
basically does the same; the only difference is that is uses a heap to
determine the largest element. In particular, it proceeds as follows:
1. The array to be sorted is turned into a heap
   (first loop in the above function)
2. The first element of the array (i.e. the largest element) is
   exchanged with its last element and the size of the array is
   reduced by 1.
3. The exchange may have destroyed the heap structure, so it is
   re-established.
4. Steps 2 and 3 are repeated until the remaining array contains
   only one element.
The main work is done by the function `sift', which (re-)establishes
the heap structure of a section of a array. It takes the first element
of an array section and lets it sift down in the heap. It does so by
checking whether the heap condition holds for this element and its
successors, i.e. whether actually both successors (provided they exist)
are smaller than the sift element. If they are or if there are no
successors, the sift element has reached its correct position and the
function returns. If they are not, the greater of the successors is
exchanged with the sift element and the process is repeated.
  The `sift' function is used in the first loop of the `heapsort'
function to turn the whole array into a heap. It is obvious that the
second half of the array (starting at index n/2) already has heap
structure, since none of these elements is the successor of any other
in this section and hence the heap condition must hold. But for the
indices n/2-1 down to 0 the heap condition may not hold, so each of
these elements is processed with the `sift' function to move it to its
proper place. After this is done, the whole array has heap structure.
  Then the first element is exchanged with the last (just as in the
selection sort algorithm), thus placing this element into the correct
position (w.r.t. sorting), and the number of elements in the unsorted
section of the array is reduced by one. Unfortunately, due to the
exchange, this section may have lost its heap structure. However, the
only element that may be in a wrong place is the one swapped to front,
so it is moved to its proper place by a call to the `sift' function.
These steps are then repeated until the heap has been reduced to one
element, which consequently must be the smallest and thus is in its
correct position (w.r.t. sorting).
  Consider the array [5 3 1 4 2] and trace how it is sorted. First
the array is turned into a heap. We mark the dividing line between
the part of the array that already has heap structure and the one
still to be processed by a '|'.

   state of the array     next operation
   5 3|1 4 2               let 3 sift down
   5|4 1 3 2               let 5 sift down
   5 4 1 3 2

Then the array is sorted by repeatedly selecting the largest element
and swapping it to the back of the unsorted section of the array (the
heap). We mark the dividing line between the sorted section (right part
of the array) and the heap (left part of the array) with a `|'.

   state of the array     next operation
   5 4 1 3 2|              exchange 2 and 5
   2 4 1 3|5               let 2 sift down
   4 3 1 2|5               exchange 2 and 4
   2 3 1|4 5               let 2 sift down
   3 2 1|4 5               exchange 3 and 1
   1 2|3 4 5               let 1 sift down
   2 1|3 4 5               exchange 2 and 1
   1|2 3 4 5

Exercise:
What is the time complexity of the heapsort algorithm?
----------------------------------------------------------------------*/

static const size_t leos[] = {  /* list of Leonardo numbers */
  1U, 1U, 3U, 5U, 9U, 15U, 25U, 41U, 67U, 109U, 177U, 287U, 465U, 753U,
  1219U, 1973U, 3193U, 5167U, 8361U, 13529U, 21891U, 35421U, 57313U,
  92735U, 150049U, 242785U, 392835U, 635621U, 1028457U, 1664079U,
  2692537U, 4356617U, 7049155U, 11405773U, 18454929U, 29860703U,
  48315633U, 78176337U, 126491971U, 204668309U, 331160281U, 535828591U,
  866988873U, 1402817465U, 2269806339U, 3672623805U,
  #if SIZE_MAX > 0xffffffff
  3672623805U, 5942430145U, 9615053951U, 15557484097U, 25172538049U,
  40730022147U, 65902560197U, 106632582345U, 172535142543U,
  279167724889U, 451702867433U, 730870592323U, 1182573459757U,
  1913444052081U, 3096017511839U, 5009461563921U, 8105479075761U,
  13114940639683U, 21220419715445U, 34335360355129U, 55555780070575U,
  89891140425705U, 145446920496281U, 235338060921987U, 380784981418269U,
  616123042340257U, 996908023758527U, 1613031066098785U,
  2609939089857313U, 4222970155956099U, 6832909245813413U,
  11055879401769513U, 17888788647582927U, 28944668049352441U,
  46833456696935369U, 75778124746287811U, 122611581443223181U,
  198389706189510993U, 321001287632734175U, 519390993822245169U,
  840392281454979345U, 1359783275277224515U, 2200175556732203861U,
  3559958832009428377U, 5760134388741632239U, 9320093220751060617U,
  15080227609492692857U,
  #endif
};

#define LEOCNT  (sizeof(leos)/sizeof(*leos))

/*--------------------------------------------------------------------*/

static void leosift (float *c, int n)
{                               /* --- let element sift down in tree */
  float *l, *r;                 /* left and right child */
  float x;                      /* buffer for sift element */

  x = *c;                       /* note the sift element */
  while (n > 1) {               /* while not at a leaf node */
    r = c -1;                   /* get right and left child */
    l = r -leos[n-2];           /* of the current tree node */
    if (*l > *r)  n -= 1;       /* determine which is larger and */
    else { l = r; n -= 2; }     /* get corresponding subtree size */
    if (x >= *l)                /* if larger child is not larger, */
      break;                    /* final position has been reached */
    *c = *l; c = l;             /* otherwise let child ascend in heap */
  }
  *c = x;                       /* store the sift element */
}  /* leosift() */

/*--------------------------------------------------------------------*/

static void trickle (float *c, int *top)
{                               /* --- let elem. trickle down in heap */
  int   n;                      /* identifier of current tree */
  float *p;                     /* child nodes and preceding root */
  float t, x;                   /* buffers for elements */

  if (top[-1] < 0) {            /* if at last tree in the heap, */
    leosift(c, *top); return; } /* merely let the root sift down */
  x = *c;                       /* note the trickle element */
  do {                          /* trickle loop */
    t = x; n = *top;            /* get value of current root */
    if (n > 1) {                /* if current root has children, */
      p  = c-1;                 /* check right and left child */
      if (*p > t) t = *p;       /* of the current root node */
      p -= leos[n-2];           /* and get largest element among */
      if (*p > t) t = *p;       /* the root and its children */
    }                           /* (to compare to preceding root) */
    p = c -leos[n];             /* get root of the preceding tree */
    if (*p <= t) break;         /* if at root of correct heap, abort */
    *c = *p;                    /* move root of preceding tree up */
    c  = p;                     /* and go to the preceding tree */
  } while (*--top >= 0);        /* while there is another tree */
  *c = x;                       /* store the trickle element and */
  leosift(c, n);                /* let it sift down in found tree */
}  /* trickle() */

/*--------------------------------------------------------------------*/

void smoothsort (float a[], size_t n)
{                               /* --- sort numbers with smoothsort */
  size_t i;                     /* loop variable */
  int    trees[LEOCNT+1];       /* list of current tree size ids */
  int    *top, k;               /* identifier of current tree */

  if (--n <= 0) return;         /* do not sort less than two elements */
  top = trees;                  /* get the tree size array */
  *top = -1; *++top = 1;        /* first element is a tree of type 1 */
  for (i = 1; i <= n; i++) {    /* add elements to Leonardo heap */
    if (top[-1] -*top == 1) {   /* if to add a root to join trees, */
      *--top += 1;              /* compute identifier of new tree */
      if (n-i <= leos[*top-1])  /* if tree root is in final heap, */
        trickle(a+i, top);      /* let root trickle down over roots, */
      else leosift(a+i, *top);} /* otherwise let root sift down */
    else if (*top != 1) {       /* if tree 1 is not present */
      *++top = 1;               /* add new element as tree 1 */
      if ((n-i <= 0)            /* if the tree will not be merged */
      || ((n-i <= 1) && (top[-1] != 2)))
        trickle(a+i, top); }    /* let root trickle down over roots */
    else {                      /* if tree 1 is already present */
      *++top = 0;               /* add new element as tree 0 */
      if (n-i <= 0)             /* if tree 0 and 1 will not be merged */
        trickle(a+i, top);      /* let root trickle down over roots */
    }                           /* (trickling is done only over roots */
  }                             /*  of trees that are in final heap) */
  for (i = n; i-- > 0; ) {      /* iteratively remove largest element */
    if (*top <= 1) {            /* if element is a single node tree, */
      top--; continue; }        /* simply remove the whole tree */
    k = *top -= 1;              /* expose the two children */
    *++top = k -= 1;            /* of the current root */
    trickle(a+i-leos[k], top-1);
    trickle(a+i,         top);  /* let roots of exposed trees */
  }                             /* trickle down over roots */
}  /* smoothsort() */

/*----------------------------------------------------------------------
Reference: E.W. Dijkstra
           Smoothsort - An Alternative for Sorting in situ
           Science of Computer Programming 1(3):223-233, 1982
Extensive explanation of the algorithm:
http://www.keithschwarz.com/smoothsort/
----------------------------------------------------------------------*/

void quicksort (float a[], size_t n)
{                               /* --- quicksort (non-recursive) */
  size_t stack[64];             /* stack for section boundaries */
  size_t *sp = stack;           /* stack pointer */
  size_t lft, rgt;              /* boundaries of array section */
  size_t l, r;                  /* indices of exchange positions */
  float  x, t;                  /* pivot element and exchange buffer */
  size_t m;                     /* num. of elements in right section */

  if (n <= 1) return;           /* do not sort less than two elements */
  *sp++ = 0;                    /* push left and right section */
  *sp++ = n -1;                 /* boundary onto the stack */
  while (sp > stack) {          /* while the stack is not empty */
    rgt = *--sp; lft = *--sp;   /* get next section boundaries */
    n = rgt -lft +1;            /* and compute number of elements */
    do {                        /* sort a section of the array */
      l = lft -1; r = rgt +1;   /* start at left and right boundary */
      x = a[lft +(n >> 1)];     /* get the middle element as pivot */
      while (1) {               /* split and exchange loop */
        while (a[++l] < x)      /* skip left  elements that are */
          ;                     /* less    than the pivot element */
        while (a[--r] > x)      /* skip right elements that are */
          ;                     /* greater than the pivot element */
        if (l >= r) {           /* if less than two elements left, */
          if (l <= r) { l++; r--; } break; }     /* abort the loop */
        t = a[l]; a[l] = a[r]; a[r] = t;
      }                         /* otherwise exchange elements */
      m = rgt -l +1;            /* compute the number of elements */
      n = r -lft +1;            /* right and left of the split */
      if (m > n) {              /* if the right section is larger, */
        if (m > 1) {            /* push its boundaries onto the stack */
          *sp++ = l; *sp++ = rgt; }                 /* (if necessary) */
        rgt = r; }              /* switch to the left section */
      else {                    /* if the left section is larger, */
        if (n > 1) {            /* push its boundaries onto the stack */
          *sp++ = lft; *sp++ = r; }                 /* (if necessary) */
        lft = l; n = m;         /* switch to the right section */
      }                         /* keeping its size m in variable n */
    } while (n > 1);            /* while there are at least */
  }                             /* two elements in the section */
}  /* quicksort() */

/*----------------------------------------------------------------------
Reference: C.A.R. Hoare
           Quicksort
           Computer Journal, 5(1):10-15, 1962

The quicksort algorithm is based on a rather simple divide-and-conquer
approach, which works as follows:
1. Select a random object (a random number) from the array to sort.
   This object (number) is called the pivot.
2. Group together all objects (numbers) that are less than the pivot
   (usually at the front of the array), and all objects (numbers) that
   are greater than the pivot (usually at the back of the array).
3. Sort the two groups recursively using the same approach.
   The recursion stops, if a group contains less than 2 objects.
The grouping in step 2 is carried out in the following way: Two
pointers, one starting at the front, the other at the back of the
array are moved towards each other. As long as the object under the
left pointer is less than the pivot, the pointer is moved right, and as
long as the object under the right pointer is greater than the pivot,
the pointer is moved left. This is done, because these objects are
already in the right position. If moving the pointers as described
above leads to a situation, in which the object under the left pointer
is greater or equal to the pivot, and the object under the right pointer
is less or equal to the pivot, the two objects are exchanged. Then
the pointers are moved again according to the above rules. When the
pointers reached or passed each other, the grouping is complete. All
objects to the left and including the position of the right pointer
are less than or equal to the pivot, and all objects to the right and
including the position of the left pointer are greater than or equal to
the pivot. Hence the boundaries of the array and the positions of the
two pointers specify the subsections to be sorted in the next step.
  To keep track of the sections that still have to be sorted, one can
use a recursive implementation (this approach is taken in the `combsort'
function below, which is essentially a combination of quicksort and
insertion sort) or one can use a stack onto which the boundaries of
one of the two sections are pushed while the other section is sorted.
The above function uses a stack. It always pushes the boundaries of
the larger section onto the stack and sorts the smaller section right
away, thus limiting the necessary size of the stack. Without this
precaution, the stack may grow in the worst case to hold the boundaries
of n/2-2 sections, where n is the number of objects in the array.
  Consider the array [5 3 1 4 2] and trace how it is sorted. First
the array is grouped according to the pivot 4, which leads to the
subsections [3 2 1] and [4 5]. The right section is already sorted,
so no exchanges are needed. Then the left section is grouped according
to the pivot 2, which leads to an exchange of 3 and 1 after which the
array is sorted.

step   state of the array
1)     3 5 4 1 2               4 is the pivot
1a)      3 2 4 1 5             5 and 2 exchanged
1b)      3 2 1 4 5             4 and 1 exchanged
2)     3 2 1|4 5               sort right section, 4 is the pivot
2a)      3 2 1|4 5             nothing to do
3)     3 2 1|4 5               sort left  section, 2 is the pivot
3a)      1 2 3|4 5             1 and 3 exchanged

After step 3a) no further recursion occurs, since the left section
contains only the 1 and the right section only the 3 (the 2 is cut out).

Exercises:
1. What is the time complexity of the quicksort algorithm
   a) on average?
   b) in the worst case?
2. Is there any way to circumvent the worst case?
----------------------------------------------------------------------*/

void cs_rec (float a[], size_t n)
{                               /* --- recursive part of combsort */
  size_t l, r;                  /* indices of exchange positions */
  float  x, t;                  /* pivot element and exchange buffer */
  size_t m;                     /* number of elements in sections */

  do {                          /* sections sort loop */
    l = 0; r = n -1;            /* start at left and right boundary */
    x = a[n >> 1];              /* get the middle element as pivot */
    if (a[l] > a[r]) { t = a[l]; a[l] = a[r]; a[r] = t; }
    if      (x < a[l]) x = a[l];   /* compute median of three */
    else if (x > a[r]) x = a[r];   /* to find a better pivot */
    while (1) {                 /* split and exchange loop */
      while (a[++l] < x)        /* skip left  elements that are */
        ;                       /* less    than the pivot element */
      while (a[--r] > x)        /* skip right elements that are */
        ;                       /* greater than the pivot element */
      if (l >= r) {             /* if less than two elements left, */
        if (l <= r) { l++; r--; } break; }       /* abort the loop */
      t = a[l]; a[l] = a[r]; a[r] = t;
    }                           /* otherwise exchange elements */
    m = n -l;                   /* compute the number of elements */
    n = r +1;                   /* right and left of the split */
    if (n > m) {                /* if right section is smaller, */
      if (m >= TH_INSERT)       /* but larger than the threshold, */
        cs_rec(a +l, m); }      /* sort it by an recursive call */
    else {                      /* if the left section is smaller, */
      if (n >= TH_INSERT)       /* but larger than the threshold, */
        cs_rec(a, n);           /* sort it by an recursive call, */
      a += l; n = m;            /* then switch to the right section */
    }                           /* keeping its size m in variable n */
  } while (n >= TH_INSERT);     /* while greater than threshold */
}  /* cs_rec() */

/*--------------------------------------------------------------------*/

void combsort (float a[], size_t n)
{                               /* --- combsort (quicksort + inssort) */
  size_t i, k;                  /* indices in number array */
  size_t s;                     /* num. of elements in first section */
  float  t;                     /* exchange buffer */

  if (n <= 1) return;           /* do not sort less than two elements */
  if (n < TH_INSERT)            /* if less elements than threshold */
    s = n;                      /* for insertion sort, note the */
  else {                        /* number of elements, otherwise */
    cs_rec(a, n);               /* call the recursive sort function */
    s = TH_INSERT -1;           /* and get the number of elements */
  }                             /* in the first array section */
  for (i = k = 0; --s > 0; )    /* find position of smallest element */
    if (a[++k] < a[i]) i = k;   /* within the first s elements */
  t = a[i]; a[i] = a[0]; a[0] = t; /* swap it to front as a sentinel */
  for (i = 0; ++i < n; ) {      /* standard insertion sort */
    t = a[i];                   /* note the number to insert */
    for (k = i; a[--k] > t; )   /* shift right all numbers that are */
      a[k+1] = a[k];            /* greater than the one to insert */
    a[k+1] = t;                 /* and store the number to insert */
  }                             /* in the place thus found */
}  /* combsort() */

/*----------------------------------------------------------------------
Combsort is a combination of a modified quicksort and an insertion sort.
The above version is inspired by the `qsort' function provided by the
Free Software Foundation (Gnu quicksort), which was adapted to sorting
arrays of floating point numbers.
  The idea underlying the combination of the two sorting algorithms is
that the grouping idea used by quicksort is very efficient for large
arrays, but rather inefficient for small ones. Hence in the above
implementation the size of the sections down to which quicksort is
carried out is limited to TH_INSERT (TH standing for threshold). The
remainder is done by applying an insertion sort to the almost sorted
array. This is efficient, since after quicksort has completed its task,
the objects in the array are close to their final positions and thus
need not be moved far by the insertion sort algorithm.
  In addition, the quicksort version used in this algorithm computes the
median of the first, the last, and the middle element of the array to
find a good pivot (provided the array contains more than 3 elements).
This reduces the danger of running into a worst case situation.

Exercise:
Does computing the median of three elements to determine the pivot
exclude the possibility of a worst case behavior?
----------------------------------------------------------------------*/

void bin2 (float a[], size_t n, float x, float dx)
{                               /* --- 2nd recursive part of binsort */
  size_t l, r;                  /* indices of exchange positions */
  float  t;                     /* (exchange) buffer */

  for (l = 0; l < n; l++)       /* find first element that is */
    if (a[l] >  x) break;       /* greater than the cut value */
  for (r = l; ++r < n; )        /* swap lesser elements to front */
    if (a[r] <= x) { t = a[l]; a[l++] = a[r]; a[r] = t; }
  dx /= 2;                      /* halve the bin size and */
  if (x -dx >= x) return;       /* check the new bin size */
  if (l    > 1) bin2(a,    l,    x -dx, dx);  /* sort left  section */
  if (n -l > 1) bin2(a +l, n -l, x +dx, dx);  /* sort right section */
}  /* bin2() */

/*--------------------------------------------------------------------*/

void bin1 (float a[], size_t n, int emin, int emax)
{                               /* --- 1st recursive part of binsort */
  size_t l, r;                  /* indices of exchange positions */
  float  x, t;                  /* cut value and exchange buffer */
  int    e, d;                  /* exponent of cut value, buffer */

  if (emin == emax) {           /* if no exponent range given, */
    x = (float)ldexp(0.25,emin);/* compute half of bin size */
    bin2(a, n, (a[0] >= 0 ? 3 : -3) *x, x);
    return;                     /* sort numbers of a given exponent */
  }                             /* with the 2nd recursive function */
  e = d = (emin +emax +1) /2;   /* compute exponent of cut value */
  x = (float)ldexp(0.5, e);     /* and the cut value itself */
  if (emax > emin) e--;         /* compute new exponent intervals */
  else { x = -x; d--; }         /* and negate cut value, if nec. */
  for (l = 0; l < n; l++)       /* find first element that is */
    if (a[l] >= x) break;       /* not less than the cut value */
  for (r = l; ++r < n; )        /* swap lesser elements to front */
    if (a[r] <  x) { t = a[l]; a[l++] = a[r]; a[r] = t; }
  if (l    > 1) bin1(a,    l,    emin, e);  /* sort left  section */
  if (n -l > 1) bin1(a +l, n -l, d, emax);  /* sort right section */
}  /* bin1() */

/*--------------------------------------------------------------------*/

void binsort (float a[], size_t n)
{                               /* --- binsort */
  size_t l, r;                  /* indices of exchange positions */
  float  t;                     /* (exchange) buffer */

  if (n <= 1) return;           /* do not sort less than two elements */
  for (l = 0; l < n; l++)       /* find first element that */
    if (a[l] >= 0) break;       /* is not less than zero */
  for (r = l; ++r < n; )        /* swap negative elements to front */
    if (a[r] <  0) { t = a[l]; a[l++] = a[r]; a[r] = t; }
  if (l    > 1) bin1(a,    l,    FLT_MAX_EXP, FLT_MIN_EXP);
  if (n -l > 1) bin1(a +l, n -l, FLT_MIN_EXP, FLT_MAX_EXP);
}  /* binsort() */

/*----------------------------------------------------------------------
Binsort is also called bucket sort or radix sort. The idea underlying
binsort is best explained by introducing, for the sake of explanation,
a severe restriction on the objects to be sorted: We confine ourselves
to sorting numbers in the range 0 to 99. It is obvious, that in this
case we can devise a very efficient algorithm to sort an array of
arbitrary size. We simply use a array with 100 counters, all set to
zero in the beginning. Then we traverse the array to sort, use the
numbers in it as indices, and increase the corresponding counters. The
sorted array is then constructed by traversing the counter array and
placing as many zeros into the number array as the counter in field 0
states, followed by as many ones as the counter in field 1 indicates,
and so on.
  Of course, this method is directly applicable only if the range of
objects is so restricted, that we can set up a ``bucket'' for each
possible object. But if we cannot, we can still use the idea. Consider,
for example, sorting a deck of cards. First you set up four buckets for
the four suits --- clubs, spades, hearts and diamonds --- and sort the
cards into them, disregarding the order within the buckets. Then you
sort each bucket, maybe by setting up three new buckets like 2 to 5,
6 to 10, and pictures and ace. In this way you proceed, until, in the
last step, there is a bucket for each card.
  If, in each step, you use two buckets, the algorithm is very similar
to the quicksort algorithm, only that the splitting (or grouping) is not
oriented at the objects that actually are in the array, but at the set
of objects that could be in the array. For example, in the above
implementation, the first step (carried out in the function `binsort')
uses two buckets, one for negative numbers and one for positive numbers
(including zero), regardless of whether there are negative numbers in
the array or not. In the second step (carried out in the function
`bin1') we look at the range of the exponents, placing numbers with an
exponents in the lower half into one bucket and all other numbers into
the other. The range of possible exponents is halved in each step,
until the numbers can differ only in their mantissa. Then the range of
values for the mantissa is halved in each step (carried out in the
function `bin2'), until there is only one possible number per bucket.
  The advantage of the binsort algorithm is that there are always some
splits that are close to optimal. The drawback of this method is that it
cannot be programmed without knowing the objects to be sorted. The above
functions, for example, depend heavily on the fact that the objects to
sort are floating point numbers.

Exercises:
1. What is the time complexity of the binsort algorithm?
2. Modify the above implementation to sort integer numbers!
----------------------------------------------------------------------*/

void error (int code, ...)
{                               /* --- print error message */
  va_list    args;              /* list of variable arguments */
  const char *msg;              /* buffer for error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  msg = errmsgs[-code];         /* get error message */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  fprintf(stderr, "\n%s: ", prgname);  /* print program name */
  va_start(args, code);         /* get variable arguments */
  vfprintf(stderr, msg, args);  /* print error message */
  va_end(args);                 /* end variable argument evaluation */
  exit(code);                   /* abort programm */
}  /* error() */

/*--------------------------------------------------------------------*/

int main (int argc, char *argv[])
{                               /* --- main function */
  size_t  i;                    /* loop variable */
  float   f;                    /* to read the numbers */
  float   *nums = NULL;         /* array of numbers to sort */
  size_t  size  = 0;            /* current size of the array */
  size_t  cnt   = 0;            /* number of numbers in the array */
  FILE    *in   = stdin;        /* file to read numbers from */
  FILE    *out  = stdout;       /* file to write sorted numbers to */
  int     method;               /* identifier of sort method */
  clock_t t;                    /* time when sort was started */

  prgname = argv[0];            /* program name for error messages */

  /* --- print a usage message --- */
  if ((argc != 3) && (argc != 4)) { /* if wrong number of args. given */
    printf("usage: %s m infile [outfile]\n", argv[0]);
    printf("sort the floating point numbers in the file "
           "'infile' with method m\n");
    printf("m = 0: qsort (stdlib)\n");
    printf("m = 1: bubblesort         m =  6: heapsort\n");
    printf("m = 2: selection sort     m =  7: smoothsort\n");
    printf("m = 3: insertion sort     m =  8: quicksort\n");
    printf("m = 4: shellsort          m =  9: combsort\n");
    printf("m = 5: mergesort          m = 10: binsort\n");
    printf("if 'infile'  starts with a '-', "
           "input is read from stdin.\n");
    printf("if 'outfile' starts with a '-', "
           "output is written to stdout.\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the function */

  /* --- get sort method id --- */
  method = atoi(argv[1]);       /* get and check the sort method id */
  if ((method < 0) || (method > 10)) error(E_METHOD, method);

  /* --- read numbers to sort --- */
  if (argv[2][0] != '-') {      /* if to read numbers from a file */
    in = fopen(argv[2], "r");   /* open the input file */
    if (!in) error(E_FOPEN, argv[2]);
    fprintf(stderr, "reading %s ... ", argv[2]);
  }
  while (fscanf(in, "%f", &f) == 1) { /* read next number */
    if (cnt >= size) {          /* if the number array is full */
      size += (size > BLKSIZE) ? (size >> 1) : BLKSIZE;
      nums = (float*)realloc(nums, size *sizeof(float));
      if (!nums) error(E_NOMEM);
    }                           /* allocate a larger number array */
    nums[cnt++] = f;            /* add the number read at the */
  }                             /* end of the number array */
  if (ferror(in)) error(E_FREAD, argv[2]);
  if (in != stdin) {            /* if input has been read from a file */
    fclose(in); fprintf(stderr, "done.\n"); } /* close the input file */
  if (cnt <= 0) {               /* check number of numbers read */
    fprintf(stderr, "%s: nothing to do\n", argv[0]); return 0; }

  /* --- sort numbers --- */
  fprintf(stderr, "sorting %"SIZE_FMT" numbers with ", cnt);
  t = clock();                  /* get the time in clock ticks */
  switch (method) {             /* evaluate the sort method id */
    case  1: fprintf(stderr, "bubblesort ... ");
             bubblesort(nums, cnt);                   break;
    case  2: fprintf(stderr, "selsort ... ");
             selsort   (nums, cnt);                   break;
    case  3: fprintf(stderr, "inssort ... ");
             inssort   (nums, cnt);                   break;
    case  4: fprintf(stderr, "shellsort ... ");
             shellsort (nums, cnt);                   break;
    case  5: fprintf(stderr, "mergesort ... ");
             mergesort (nums, cnt);                   break;
    case  6: fprintf(stderr, "heapsort ... ");
             heapsort  (nums, cnt);                   break;
    case  7: fprintf(stderr, "smoothsort ... ");
             smoothsort(nums, cnt);                   break;
    case  8: fprintf(stderr, "quicksort ... ");
             quicksort (nums, cnt);                   break;
    case  9: fprintf(stderr, "combsort ... ");
             combsort  (nums, cnt);                   break;
    case 10: fprintf(stderr, "binsort ... ");
             binsort   (nums, cnt);                   break;
    default: fprintf(stderr, "stdlib qsort ... ");
             qsort(nums, cnt, sizeof(float), fltcmp); break;
  }                             /* sort the number array and */
  t = clock() -t;               /* compute the execution time */
  fprintf(stderr, "[%.2fs] done.\n", (double)t/CLOCKS_PER_SEC);

  #if CHECK                     /* if to check the result */
  for (i = cnt; --i > 0; )      /* traverse the number array */
    if (nums[i-1] > nums[i]) {  /* and check the number order */
      fprintf(stderr, "%s: sort error\n", argv[0]); return -1; }
  fprintf(stderr, "%s: sort ok\n", argv[0]);
  #endif

  /* --- print the sorted numbers --- */
  if (argc > 3) {               /* if output file name given */
    if (argv[3][0] != '-') {    /* if to write numbers to a file */
      out = fopen(argv[3], "w");/* open the output file */
      if (!out) error(E_FOPEN, argv[3]);
      fprintf(stderr, "writing %s ... ", argv[3]);
    }
    for (i = 0; i < cnt; i++)   /* traverse the sorted array and */
      fprintf(out, "%g\n", nums[i]);         /* print the numbers */
    if (out != stdout) {        /* if output was to a file */
      if (fclose(out) != 0) error(E_FWRITE, argv[3]);
      fprintf(stderr,"done.\n");/* close the output file and */
    }                           /* print a success message */
  }
  free(nums);                   /* delete the number array */
  return 0;                     /* return 'ok' */
}  /* main() */
